/*
    Physics Simulation

    (C) 2007 Jason Hunt
    nulluser@gmail.com

    File: physics.cpp
*/

#include "core.h"


/* Render object */
void dobject_class::render( void )
{
    // Set active texture
    world.dx_device->SetTexture(0, world.dx_texture2);

    // Copy from vertex buffer
    world.dx_device->SetStreamSource(0, dx_vbuff, 0, sizeof(vertex));

    // Draw normally
    if (!world.drawinternal)
    {
        // Copy new position data into vertex buffer

        if (copy_face_data()) return;    

        // Draw the primitives
        if (world.dx_device->DrawPrimitive(D3DPT_TRIANGLELIST, 0, num_triangles) != D3D_OK)
           return;
    }

    // Draw internal structure
    if (world.drawinternal)
    {
        if (!copy_line_data()) return;

        // Draw the primitives
        if (world.dx_device->DrawPrimitive(D3DPT_LINELIST, 0, num_lines) != D3D_OK)
            return;;
    }   
}
/* End of  dobject_class::render */




/* Get a tension from a delta and a relaxed length */
void dobject_class::get_tension( tension_type &t, state delta, double base )
{
    t.delta = delta.pos;

    double len = delta.pos.mag();        // Length of displacement

    if (len < epsilon) len = epsilon;

    t.norm_delta = delta.pos / len;


    if (world.apply_snap)
        if (len > base * snap_percent || len < base / snap_percent) t.snapped = true;
    
    // No force if snapped
    if (t.snapped) 
    {
        t.force = zero;
        return;
    }

    // Calculate force using hooks law
//    t.force = spring_k * (base - len)  * delta.pos / len;
    t.force = (spring_k * (base - len)  * delta.pos - dampening * delta.vel)/len;
}
/* End of get tension */


/* Get position vector from point indexes */
state dobject_class::get_state( int ix, int iy, int iz )
{
    // Return zero if out of range
    if (ix < 0 || ix >= num_points.x ||
        iy < 0 || iy >= num_points.y ||
        iz < 0 || iz >= num_points.z) return(state(zero, zero));
       
    return(points[get_index(ix, iy, iz)].cur_state);
}
/* End of get_pos */


/* Calculate all tensions */
void dobject_class::calc_tensions(void)
{
    // Calculate tensions in x, y and z directions for all points
    for (int iz = 0; iz < num_points.z; iz++)
        for (int iy = 0; iy < num_points.y; iy++)
            for (int ix = 0; ix < num_points.x; ix++)
            {
                int i = get_index(ix, iy, iz);

                state base = points[i].cur_state;
                
                get_tension(tension_x[i], get_state(ix+1,iy,iz)-base, rest_len.x);     
                get_tension(tension_y[i], get_state(ix,iy+1,iz)-base, rest_len.y);                     
                get_tension(tension_z[i], get_state(ix,iy,iz+1)-base, rest_len.z);     
            }

}
/* End of void dobject_class::calc_tensions(void) */



/* Calculate sum of forces for a point */
vector dobject_class::calc_point_force(unsigned int ix, unsigned int iy, unsigned int iz )
{
    vector f = zero;   
   
    // The following code calculates forces from spring tension
    // and angular displacement from the reverence directions

    int i;

    // Forces working on positive x axis tension
    if (ix > 0)
    {
        i = get_index(ix-1, iy, iz);
        f += tension_x[i].force;  
        f += torque_k * (unit_x - tension_x[i].norm_delta);
    }

    // Forces working on negative x axis tension
    if (ix < num_points.x - 1)
    {    
        i = get_index(ix, iy, iz);        
        f -= tension_x[i].force;  
        f -= torque_k * (unit_x - tension_x[i].norm_delta);
    }

    // Forces working on positive y axis tension
    if (iy > 0)
    {
        i = get_index(ix, iy-1, iz);
        f += tension_y[i].force;   
        f += torque_k * (unit_y - tension_y[i].norm_delta);
    }
    
    // Forces working on negative y axis tension
    if (iy < num_points.y - 1)
    {
        i = get_index(ix, iy, iz);
        f -= tension_y[i].force;  
        f -= torque_k * (unit_y - tension_y[i].norm_delta);
    }

    // Forces working on positive z axis tension
    if (iz > 0)              
    {
        i = get_index(ix, iy, iz-1);
        f += tension_z[i].force; 
        f += torque_k * (unit_z - tension_z[i].norm_delta);
    }

    // Forces working on negative z axis tension     
    if (iz < num_points.z - 1)
    {
        i = get_index(ix, iy, iz);
        f -= tension_z[i].force;   
        f -= torque_k * (unit_z - tension_z[i].norm_delta);
    }
    
    return(f);   
}
/* End of calc_point_force */


/* Claculate all forces */
void dobject_class::calc_forces( void )
{
    unsigned int i = 0;
    
    for (int iz = 0; iz < num_points.z; iz++)
        for (int iy = 0; iy < num_points.y; iy++)
            for (int ix = 0; ix < num_points.x; ix++)
            {
                //unsigned int i = get_index(ix, iy, iz);
                
                // Do not calculate forces for fixed points
                if (!points[i].fixed)
                    points[i].force = calc_point_force(ix, iy, iz);
                    i++;
                
            }
}
/* End of move objects */


/* Calculate tensions and forces */
void dobject_class::intergrate_step( void )
{
    calc_tensions();
    calc_forces();
}
/* End of void dobject_class::intergrate_step( void ) */


/* Clip a point into the display area */
void dobject_class::clip( mass_type &p )
{
    if (p.world_pos.x < world.world_min.x) { p.world_pos.x = world.world_min.x;  p.cur_state.vel.x *= -1; }
    if (p.world_pos.x > world.world_max.x) { p.world_pos.x = world.world_max.x;  p.cur_state.vel.x *= -1; }
    if (p.world_pos.y < world.world_min.y) { p.world_pos.y = world.world_min.y;  p.cur_state.vel.y *= -1; }
    if (p.world_pos.y > world.world_max.y) { p.world_pos.y = world.world_max.y;  p.cur_state.vel.y *= -1; }
    if (p.world_pos.z < world.world_min.z) { p.world_pos.z = world.world_min.z;  p.cur_state.vel.z *= -1; }
    if (p.world_pos.z > world.world_max.z) { p.world_pos.z = world.world_max.z;  p.cur_state.vel.z *= -1; }
}
/* End of  dobject_class::clip */


/* Returns true if points does not need calculation */
bool dobject_class::exclude_point(int ix, int iy, int iz)
{
    return(false);
}
/* End of dobject_class::exclude_point(int ix, int iy) */



/* Function for RK4 */
state dobject_class::f(mass_type &p)
{
    // pos' = vel and vel' = accl
    return state(p.cur_state.vel,  (p.force + p.ext_force - drag * p.cur_state.vel) / point_mass);
}
/* End of f */


/* Preform rk4 integration */
void dobject_class::intergrate_rk4( void )
{
    state s[total_points],       // Starting state
          k1[total_points],      // Temp values
          k2[total_points], 
          k3[total_points];

    // Delta t between intergrations
    double h = world.get_delta_t();
   
    // Phase 1

    // Calculate forces
    // This does not change the pos or velocity, no need to copy again
    intergrate_step();

    for (unsigned int p = 0; p < total_points; p++)
    {
        s[p] = points[p].cur_state;                 // Copy starting state
        k1[p] = h * f(points[p]);
        points[p].cur_state = s[p] + k1[p] / 2.0; 
    }

    // Phase 2

    intergrate_step();

    for (unsigned int p = 0; p < total_points; p++)
    {
        k2[p] = h * f(points[p]); 
        points[p].cur_state = s[p] + k2[p] / 2.0; 
    }

    // Phase 3

    intergrate_step();

    for (unsigned int p = 0; p < total_points; p++)
    {
        k3[p] = h * f(points[p]);
        points[p].cur_state = s[p] + k3[p]; 
    }

    // Phase 4

    intergrate_step();

    for (unsigned int p = 0; p < total_points; p++)
        points[p].cur_state = s[p] + (k1[p] + 2*(k2[p] + k3[p]) + h*f(points[p])) / 6.0;
}
/* End of dobject_class::intergrate_rk4 */




/* Update reference frame and local points */
void dobject_class::update_state(void)
{   
    state ave_state = state(zero, zero);
 
    // Use the same time step as RK4 integrate
    double dt = ( world.slowmode ? world.rk4_delta_slow : world.rk4_delta_fast) ; 
   
   // Compute average states
    for (unsigned int p = 0; p < total_points; p++)
        ave_state += points[p].cur_state;

    ave_state = ave_state / (double)total_points;

    // Subtract out global movement and position
    for (unsigned int p = 0; p < total_points; p++)
        points[p].cur_state -= ave_state;

    vector torque_sum = zero;

    // Calculate all torques
    for (unsigned int p = 0; p < total_points; p++)
    {
        // The position is relative to the center of mass
        // So it's the same as a delta from the point to the center
       
        torque_sum += cross(points[p].cur_state.pos, points[p].ext_force);
    }
    
    torque_sum = torque_sum / (double)total_points;     // Normalize over points

    ang_vel += 0.05 * torque_sum * dt;                  // Compute angular velocity
    ang_vel -= 4.40 * ang_vel * dt;                     // Dampening on angular velocity
  
    
    // Update and renormalise the orientation
    orientation += orientation * ang_vel;

    orientation.normalize();
    
    // Need to rotate pand translate into worldspace
    for (unsigned int p = 0; p < total_points; p++)
    {
        // Copy position
        points[p].world_pos = points[p].cur_state.pos;
        
        // Rotate into world space
        quaternion_rotate_vector(orientation, points[p].world_pos);

        // Translate into world space 
        points[p].world_pos += cur_state.pos;
    }
  
    // Delta velocity has to be rotated into world space
    quaternion_rotate_vector(orientation,  ave_state.vel);
   
    // Update position and velocity of refernce frame
    cur_state.pos += ave_state.pos + cur_state.vel * dt;  

    // Add new frame velocity
    cur_state.vel += ave_state.vel;
       
    // Apply gravity
    if (world.apply_grav)
        cur_state.vel -= vector(0.0, total_mass * world.grav,0.0) * dt;
    
    // Apply air friction
    cur_state.vel -= 1.0 * cur_state.vel * dt;    
    
    // Ground friction
    if (apply_ground_friction)
        cur_state.vel -= 5.0 * cur_state.vel * dt;    
}
/* End of update_state */



/* Update physics for dobject */
void dobject_class::physics_update( void )
{
    intergrate_rk4();
    
    update_state();
}
/* End of void dobject_class::physics_update( void ) */



/* Determine which point is close to the passed world cords */
void dobject_class::get_close_point(double wx, double wy, int &ix, int &iy)
{
    /*double dist = max_x * max_y;
    ix = 1;
    iy = 0;
    
    for (int y = 0; y < points_y; y++)
        for (int x = 0; x < points_x; x++)
        {
            double d = pow(wx - points[0][y][x].pos.x, 2) + 
                        pow(wy - points[0][y][x].pos.y, 2);
                        
            if (d < dist)
            {
                dist = d;
                ix = x;
                iy = y;
            }
        }    

//    text_pt_x = x;
//    text_pt_y = y;
*/
}
/* End of get_close_point */



/* Return base point index from xyz */
int dobject_class::get_index(int ix, int iy, int iz)
{
    if (ix < 0 || ix >= num_points.x) return(0);
    if (iy < 0 || iy >= num_points.y) return(0);
    if (iz < 0 || iz >= num_points.z) return(0);

    return(iz * num_points.x * num_points.y + iy * num_points.x + ix);
}
/* End of dobject_class::get_index */

/* Inc number of point in x direction */
void dobject_class::inc_points_x ( void )
{
    num_points.x++; 
    init(); 
}
/* End of inc_points_x */

/* Dec number of point in x direction */
void dobject_class::dec_points_x ( void )
{
    if (num_points.x <= 2) return;
    num_points.x--;
    init(); 
}
/* End of dec_points_x */

/* Inc number of point in y direction */
void dobject_class::inc_points_y ( void )
{
    num_points.y++; 
    init();
}
/* End of inc_points_y */

/* Dec number of point in y direction */
void dobject_class::dec_points_y ( void )
{
    if (num_points.y <= 2) return;
    num_points.y--; 
    init(); 
}
/* End of dec_points_y */

/* Inc number of point in z direction */
void dobject_class::inc_points_z ( void )
{
    num_points.z++;
    init(); 
}
/* End of inc_points_z */

/* Dec number of point in z direction */
void dobject_class::dec_points_z ( void )
{
    if (num_points.z <= 2) return;
    num_points.z--;
    init();    
}
/* End of dec_points_z */

/* Inc mass */
void dobject_class::inc_mass ( void )
{
    total_mass *= 1.01; 
    update_const(); 
}
/* End of inc_mass */

/* Dec mass */
void dobject_class::dec_mass ( void )
{
    total_mass /= 1.01; 
    update_const(); 
}
/* End of dec_mass */  
  
  
/* Inc spring constant */
void dobject_class::inc_spring_k ( void )
{ 
    base_spring_k *= 1.01; 
    update_const(); 
}
/* End of inc_spring_k */

/* Dec spring constant */
void dobject_class::dec_spring_k ( void )
{ 
    base_spring_k /= 1.01; 
    update_const(); 
}
/* End of dec_spring_k */

/* Inc torque constant */
void dobject_class::inc_torque_k ( void )
{ 
    base_torque_k *= 1.01; 
    update_const(); 
}
/* End of inc_torque_k */

/* Dec torque constant */
void dobject_class::dec_torque_k ( void )
{ 
    base_torque_k /= 1.01; 
    update_const(); 
}
/* End of dec_torque_k */

/* Inc dragconstant */
void dobject_class::inc_drag ( void )
{ 
    base_drag *= 1.01; 
    update_const(); 
}
/* End of inc_torque_k */

/* Dec drag */
void dobject_class::dec_drag ( void )
{ 
    base_drag /= 1.01; 
    update_const(); 
}
/* End of dec_torque_k */







