/*
    Physics Simulation

    (C) 2007 Jason Hunt
    nulluser@gmail.com

    File: dobject_init.cpp
*/

#include "core.h"


/* Constructor for dynamic object class  */
dobject_class::dobject_class( void )
{
    set_base_const();
    
    init(); // first init
}
/* End of dobject Constructor */


/* Constructor for dynamic object class with a specified position and velocity */
dobject_class::dobject_class( vector p, vector v )
{
    set_base_const();
    
    cur_state.pos = p;
    cur_state.vel = v;
        
    init(); // first init
}
/* End of dobject Constructor */


/* Setup base constants */
void dobject_class::set_base_const( void )
{
    ang = vector(0.0, 0.0, 0.0);

    ang_vel = zero;

    test_pt_x = 0;                      // Cords of grabbed point
    test_pt_y = 0;

    cur_state.pos = vector(f_rand(-50, 50), 10.0, f_rand(-50, 50));
    cur_state.vel = zero;

    size = vector(3.0, 3.0, 3.0);    // Size in 3d
    
    num_points = intvector(4,4,4);      // Number of points 


    snap_percent = 1.50;                // Percent spring can be streched or compressed before snap
    total_mass = 5.0;                   // Mass of entire object

    base_dampening = 20.0;

//    dampening = 0.4;

    apply_ground_friction = false;

    base_drag = 500.0;
//    base_drag = 50.0;
    
    base_spring_k = 30000;              // Base Spring Constant
    base_torque_k = 30000;              // Base Torque Constant
  
    var_base = 0.01 * size.mag();        // Base for pos variance

    total_points = 0;                   // Total number of points

    num_triangles = 0;
    num_lines  = 0;

              
    // Null all memory pointers

    points = NULL;
    vert_data = NULL;

    dx_vbuff = NULL;                    // Clear Vertex Buffer

    tension_x = tension_y = tension_z = NULL;
    
    next = NULL;    
}
/* End of set_base_const */




/* dobject destructor */
dobject_class::~dobject_class()
{
    free_memory();
}
/* End of dobject destructor */


/* Update simulation constants */
void dobject_class::update_const( void )
{
    total_points = num_points.x * num_points.y * num_points.z;

    num_triangles = 4*( (num_points.x-1) * (num_points.y-1) + 
                        (num_points.x-1) * (num_points.z-1) +
                        (num_points.y-1) * (num_points.z-1) );

    num_face_points = num_triangles * 3;
    

    point_mass = total_mass / (double)(total_points);

    spring_k = base_spring_k / (double)(total_points);    
    torque_k = base_torque_k / (double)(total_points);    
    drag = base_drag / (double)(total_points);  

    dampening = base_dampening / (double)(total_points); 

    point_size = size / (double)(total_points);
   
    rest_len.x = size.x / (double)(num_points.x - 1);
    rest_len.y = size.y / (double)(num_points.y - 1);
    rest_len.z = size.z / (double)(num_points.z - 1);
                                     
    mass_radius = (rest_len.x + rest_len.y + rest_len.z) / 3.0;
     
    // inital     orientation
    orientation = quaternion_from_angles(ang);
}
/* End of update_const */



/* Free all object memory */
void dobject_class::free_memory(void)
{
    if (tension_x)  delete [] tension_x;
    if (tension_y)  delete [] tension_y;
    if (tension_z)  delete [] tension_z;
    
    if (points) delete [] points;
  
    if (face_masses) delete [] face_masses;
  
    if (vert_data) delete [] vert_data;
    
    // Free old vertex buffer data
    if (dx_vbuff) dx_vbuff->Release();    
}
/* End of free memory */


/* Allocate memory for object */
void dobject_class::get_memory(void)
{
    free_memory();                              // Free old memory 
       
    points = new mass_type[total_points];
    
    face_masses = new face_mass_type[num_face_points];
    
    tension_x = new tension_type[total_points];
    tension_y = new tension_type[total_points];
    tension_z = new tension_type[total_points];

    vert_data = new vertex[num_face_points];

    // Get vertex buffer
    if (world.dx_device->CreateVertexBuffer(num_face_points * sizeof(vertex),
                                            0, dx_fvf, D3DPOOL_MANAGED,
                                            &dx_vbuff, NULL) != D3D_OK)
    {
        debug("Unable to create vertex buffer");
        running = false;
    }       
}
/* End of get_memory */



/* Reset object variables */
void dobject_class::init( void )
{
    update_const();     // (Re)calculate constants

    get_memory();       // (Re)alloc memory

    // Clear snaps for tensions
    for (int i = 0; i < total_points; i++)
    {
        tension_x[i].snapped = false;
        tension_y[i].snapped = false;
        tension_z[i].snapped = false;
    }        
           
        
    // Calculate variances
    double vx = var_base/(double)(num_points.x);
    double vy = var_base/(double)(num_points.y);
    double vz = var_base/(double)(num_points.z);


    // Setup inital positions             
    for (int z = 0; z < num_points.z; z++)
    {
        double bz = size.z * (z/(double)(num_points.z - 1) - 0.5);

        // Setup inital positions             
        for (int y = 0; y < num_points.y; y++)
        {
            double by = size.y * (y/(double)(num_points.y - 1) - 0.5);

            for (int x = 0; x < num_points.x; x++)
            {
                double bx = size.x * (x/(double)(num_points.x - 1) - 0.5);
                      
                mass_type &p = points[get_index(x,y,z)];
                       
                p.cur_state.pos = vector(bx, by, bz);
                
                p.cur_state.pos.x += f_rand(-vx, vx);
                p.cur_state.pos.y += f_rand(-vy, vy);
                p.cur_state.pos.z += f_rand(-vz, vz);
                
//                p.base_pos =  p.state.pos;
                           
                //p.abs_pos = p.cur_state.pos + cur_state.pos;
                   
                   
                                              
                p.cur_state.vel = zero;//0;//vel;
//                p.accl = zero;

//                if (y == 1)//num_points.y -1)           
//                    p.fixed = true;
//                else
                    p.fixed = false;     


                p.force = zero;
                p.ext_force = zero;
                    
                    
//            if (x == num_points.x -1 && y == num_points.y -1 && z == num_points.z -1)
//               p.fixed = true;  


          // if (x == 0 && y == num_points.y -1 && z == 0)
          //     p.fixed = true;  
                                            
            }
        
        }
    }
    
    setup_face_masses(); // create face list

    update_state();
}
/* End of init */



/* Add face mass to list */
void dobject_class::set_face_mass(unsigned int &i, mass_type *mass_ptr, float tx_x, float tx_y)
{
    face_masses[i].mass = mass_ptr;
    face_masses[i].tx_x = tx_x;
    face_masses[i].tx_y = tx_y;
    i++;
}
/* End of set_face_mass */




/* Copy faces on the X-Y plane */
void dobject_class::make_xy_faces( unsigned int &i, float sub_x, float sub_y, float sub_z)
{
  // Copy front and rear triangle data x-y plane
    for (int y = 0; y < num_points.y-1; y++)
    {
        float tx_y = (num_points.y > 1) ? y / (float)(num_points.y-1) : 0;

        for (int x = 0; x < num_points.x-1; x++)
        { 
            float tx_x = (num_points.x > 1) ? x / (float)(num_points.x-1) : 0;

    
            float tx_x1 = tx_x + sub_x;
            float tx_y1 = tx_y + sub_y;            
                
            // top triangle front face
            set_face_mass(i, &points[get_index(x,y,0)], tx_x,  tx_y);                            
            set_face_mass(i, &points[get_index(x+1,y+1,0)], tx_x1, tx_y1); 
            set_face_mass(i, &points[get_index(x+1,y,0)],   tx_x1, tx_y); 

            // bottom triangle front face
            set_face_mass(i, &points[get_index(x+1,y+1,0)], tx_x+sub_x,tx_y+sub_y);
            set_face_mass(i, &points[get_index(x,y,0)], tx_x, tx_y); 
            set_face_mass(i, &points[get_index(x,y+1,0)], tx_x,tx_y+sub_y); 

            // top triangle back face
            set_face_mass(i, &points[get_index(x,y,num_points.z-1)], tx_x,  tx_y);                            
            set_face_mass(i, &points[get_index(x+1,y,num_points.z-1)], tx_x+sub_x   ,tx_y); 
            set_face_mass(i, &points[get_index(x+1,y+1,num_points.z-1)], tx_x +sub_x  ,tx_y+sub_y); 

            // bottom triangle front face
            set_face_mass(i, &points[get_index(x+1,y+1,num_points.z-1)], tx_x+sub_x,tx_y+sub_y);
            set_face_mass(i, &points[get_index(x,y+1,num_points.z-1)], tx_x,tx_y+sub_y); 
            set_face_mass(i, &points[get_index(x,y,num_points.z-1)], tx_x, tx_y); 
        }
    }
    
}
/* End of make_xy_faces */

/* Copy faces on the X-Z plane */
void dobject_class::make_xz_faces( unsigned int &i, float sub_x, float sub_y, float sub_z)
{

    // Top and bottom of square  in x-z plane
    for (int z = 0; z < num_points.z-1; z++)
    {
        float tx_z = z / (float)(num_points.z-1);

        for (int x = 0; x < num_points.x-1; x++)
        { 
            float tx_x = x / (float)(num_points.x-1);

    
            set_face_mass(i, &points[get_index(x,num_points.y-1,z)], tx_x,  tx_z);                            
            set_face_mass(i, &points[get_index(x+1,num_points.y-1,z+1)], tx_x +sub_x  ,tx_z+sub_z); 
            set_face_mass(i, &points[get_index(x+1,num_points.y-1,z)], tx_x+sub_x   ,tx_z); 

            set_face_mass(i, &points[get_index(x,num_points.y-1,z)], tx_x,  tx_z);                            
            set_face_mass(i, &points[get_index(x,num_points.y-1,z+1)], tx_x   ,tx_z+sub_z); 
            set_face_mass(i, &points[get_index(x+1,num_points.y-1,z+1)], tx_x+sub_x   ,tx_z+sub_z); 

            set_face_mass(i, &points[get_index(x,0,z)], tx_x,  tx_z);                            
            set_face_mass(i, &points[get_index(x+1,0,z)], tx_x+sub_x   ,tx_z); 
            set_face_mass(i, &points[get_index(x+1,0,z+1)], tx_x +sub_x  ,tx_z+sub_z); 

            set_face_mass(i, &points[get_index(x,0,z)], tx_x,  tx_z); 
            set_face_mass(i, &points[get_index(x+1,0,z+1)], tx_x+sub_x   ,tx_z+sub_z); 
            set_face_mass(i, &points[get_index(x,0,z+1)], tx_x   ,tx_z+sub_z); 
        }
    }        
}
/* End of make_xz_faces */

/* Copy faces on the Y-Z plane */
void dobject_class::make_yz_faces( unsigned int &i, float sub_x, float sub_y, float sub_z)
{

    // Top and bottom of square  in y-z plane
    for (int z = 0; z < num_points.z-1; z++)
    {
        float tx_z = z / (float)(num_points.z-1);        

        for (int y = 0; y < num_points.y-1; y++)
        { 
            float tx_z = z / (float)(num_points.z-1);
            float tx_y = y / (float)(num_points.y-1);
    
            set_face_mass(i, &points[get_index(0,y,z)], tx_z,  tx_y);                            
            set_face_mass(i, &points[get_index(0,y+1,z+1)], tx_z +sub_z  ,tx_y+sub_y); 
            set_face_mass(i, &points[get_index(0,y+1,z)], tx_z  ,tx_y+sub_y); 

            set_face_mass(i, &points[get_index(0,y,z)], tx_z,  tx_y);                            
            set_face_mass(i, &points[get_index(0,y,z+1)], tx_z+sub_z   ,tx_y); 
            set_face_mass(i, &points[get_index(0,y+1,z+1)], tx_z +sub_z  ,tx_y+sub_y); 

            set_face_mass(i, &points[get_index(num_points.x-1,y,z)], tx_z,  tx_y);                            
            set_face_mass(i, &points[get_index(num_points.x-1,y+1,z)], tx_z   ,tx_y+sub_y); 
            set_face_mass(i, &points[get_index(num_points.x-1,y+1,z+1)], tx_z +sub_z  ,tx_y+sub_y); 

            set_face_mass(i, &points[get_index(num_points.x-1,y,z)], tx_z,  tx_y);                            
            set_face_mass(i, &points[get_index(num_points.x-1,y+1,z+1)], tx_z +sub_z  ,tx_y+sub_y); 
            set_face_mass(i, &points[get_index(num_points.x-1,y,z+1)], tx_z+sub_z   ,tx_y); 
        }  
    }
    
}
/* End of make_yz_faces */


void dobject_class::setup_face_masses( void )
{
  
    // Calculate width of sub texture
    float sub_x = (num_points.x>1) ? 1 / (float)(num_points.x-1) : 0;
    float sub_y = (num_points.y>1) ? 1 / (float)(num_points.y-1) : 0;
    float sub_z = (num_points.z>1) ? 1 / (float)(num_points.z-1) : 0;
       
    unsigned int i= 0;
     
    make_xy_faces(i, sub_x, sub_y, sub_z);
    make_xz_faces(i, sub_x, sub_y, sub_z);
    make_yz_faces(i, sub_x, sub_y, sub_z);
}




