/*
    Physics Simulation

    (C) 2007 Jason Hunt
    nulluser@gmail.com

    File: world.cpp
*/


#include "core.h"


world_class world;      // Entire simulation


/* World constructor, setup base parameters */
world_class::world_class( void )
{
    debug_start();

    display_x = 800;                    // Screen size
    display_y = 600;

//    display_x = 1280;                    // Screen size
//    display_y = 1024;

    // World size in meters
    world_min = vector(-100.0, 0.0, -100.00);  
    world_max = vector(100.0, 50.0, 100.0);

    frames = 0;                         // Number of frames in this 250ms
    fps = 0;                            // Frames per second

    fullscreen = false;                 // True if in fullscreen mode

    wireframe = false;
    drawinternal = false;

    cam_speed = 0.065;    

    camera_pos =  D3DXVECTOR3(0.0,4.0,-10.0);
    camera_dir = D3DXVECTOR3(0.0,0.0,1.0);

    camera_ang = vector(0.0,0.0,0.0);

    dx_font;                                    // Test font
    dx_texture1 = NULL;                         // Test texture
    dx_texture2 = NULL;                         // Test texture

    dx_white = D3DCOLOR_ARGB(255, 255, 255, 255);
    dx_black = D3DCOLOR_ARGB(0, 0, 0, 0);


    terrain_x_cells = 20;
    terrain_z_cells = 20;
    terrain_triangles = 0;    
 
    terrain_buff = NULL;                    // Vertex Buffer   

    paused = false;                         // True if simulation is stopped

    apply_grav = true;                      // True if gravity is enabled

    apply_snap = false;                     // True if breaking is enabled

    slowmode = false;                      // True if simulation is in slow mode

    //double base_drag = 0.40;              // Drag
    grav = 9.8;                          // Gravity in meters / second
    rk4_delta_fast = 0.01;               // Delta time for RK4 integration
    rk4_delta_slow = 0.001;              // Delta time for RK4 integration

    terrain_force = 1000.0;
    camera_force = 100.0;
    object_force = 1000.0;

    probe_dist = 3.0;

    probe_radius = 1.5;
    
    num_dobjects = 0;
}
/* End of world class constructor */


/* Setup simulation */
int world_class::start(HWND hwnd)
{
    srand(GetTickCount());                  // Seed the random number generator

    if (start_graphics(hwnd)) return(1);

   // for (int i  = 0; i < 100; i++)    
    add_projectile();

    return(0);
}
/* End of start */
    
    
/* Free all world data */
void world_class::stop(void)
{
    stop_graphics();
  
    free_objects();
}
/*End of world_class::stop */
      

/* Create ground terrain */
void world_class::create_terrain( void )
{
    debug("world_class::create_terrain");

    terrain_triangles = terrain_x_cells * terrain_z_cells * 2;

    // Get vertex buffer
    if (dx_device->CreateVertexBuffer(terrain_triangles * 3 * sizeof(vertex),
                                      0, dx_fvf,
                                      D3DPOOL_MANAGED,
                                      &terrain_buff, NULL) != D3D_OK) return;

    vector tmp_points[terrain_z_cells+1][terrain_x_cells+1];

    double disp = 0;//1.0;

    for (int z = 0; z < terrain_z_cells+1; z++)
    {
        double rz = z/(double)terrain_z_cells * (world_max.z - world_min.z) + world_min.z;
        
        for (int x = 0; x < terrain_x_cells+1; x++)
        {
            double rx = x/(double)terrain_x_cells * (world_max.x - world_min.x) + world_min.x;
            
            tmp_points[z][x] = vector(rx, f_rand(-disp,disp),rz);
        }
    }

    int i = 0;

    vertex *tmp_data = new vertex[terrain_triangles*3 ];

    for (int z = 0; z < terrain_z_cells; z++)
    {
        for (int x = 0; x < terrain_x_cells; x++)
        {
            vector p1 = tmp_points[z][x];
            vector p2 = tmp_points[z][x+1];
            vector p3 = tmp_points[z+1][x];
            vector p4 = tmp_points[z+1][x+1];
                                    
            tmp_data[i++] = vertex(p1, 0,  0);                            
            tmp_data[i++] = vertex(p3, 0,  1);                            
            tmp_data[i++] = vertex(p4, 1,  1);                            

            tmp_data[i++] = vertex(p4, 1,  1);                            
            tmp_data[i++] = vertex(p2, 1,  0);                            
            tmp_data[i++] = vertex(p1, 0,  0);                            

        }
    }

    VOID *data; 

    // Try to lock the vertex buffer
    if (terrain_buff ->Lock(0, 0, (void**)&data, 0) != D3D_OK) return;
    
    // Copy data into vertext buffer
    memcpy(data, tmp_data, sizeof(vertex) * terrain_triangles  * 3);

    
    terrain_buff->Unlock(); 
    
    delete [] tmp_data;    
}
/* End of create_terrain */


void world_class::update_fps( void )
{
    static double last_time = get_time();
    
    double cur_time = get_time();
    double diff = cur_time - last_time;

    if (fabs(diff) < epsilon) return;
            
    fps = (unsigned int)(frames / diff);

    last_time = cur_time;

    frames = 0;    
}



/* Core simulation function */
void world_class::next_frame( void )
{
    world.update_camera();

    if (!world.paused)
    {
        world.check_collide();

        dobject_class *d = world.objects;
        
        while (d != NULL)
        {
            d->physics_update();
            d = d->next;
        }
    }

    world.render();
}
/* End of next frame */


/* Free all worl objects */
void world_class::free_objects(void)
{
    dobject_class *tmp;

    while (objects != NULL)
    {
        tmp = objects->next;

        if (objects)
            delete objects;

        objects = tmp;
    }
    
    num_dobjects = 0;    
}
/* End of free objects */


/* Add a random object */
 void world_class::add_object(void)
{
    vector pos;

    pos.x = camera_pos.x + 5 * camera_dir.x;
    pos.y = camera_pos.y + 5 * camera_dir.y;
    pos.z = camera_pos.z + 5 * camera_dir.z;    

    dobject_class *tmp = objects;
    
    objects = new dobject_class;
    
    objects->next = tmp;
    
    objects->cur_state.pos = pos;
    
    objects->ang = camera_ang;
    
    
    objects->init();
    
    num_dobjects++;        
}
 
 
/* Add a projectile to the system */
 void world_class::add_projectile(void)
{
    vector pos;
    pos.x = camera_pos.x + 5 * camera_dir.x;
    pos.y = camera_pos.y + 5 * camera_dir.y;
    pos.z = camera_pos.z + 5 * camera_dir.z;
    
    // = camera_pos;
    vector vel;// = camera_dir;
    
    vel.x = 40 * camera_dir.x;
    vel.y = 40 * camera_dir.y;
    vel.z = 40 * camera_dir.z;

    dobject_class *tmp = objects;
    
    objects = new dobject_class(pos, vel);

    objects->is_projectile = true;

  //  objects->size = get_point(1.0, 1.0, 1.0);

    objects->base_spring_k = 10000;
    
    objects->ang = camera_ang;
    

    objects->init();
    
        
        
        
    objects->next = tmp;
    
    num_dobjects++;        
}
/* End of add_projectile */




